/*************************************************************
 * Rover set up to demonstrate magnetometer
 */

#include <Wire.h>
#include <Adafruit_Sensor.h>
#include <Adafruit_LSM303_U.h>
#include "vectors.h"
#include <EEPROM.h>

/* Assign a unique ID to the accel sensor */
Adafruit_LSM303_Accel_Unified acc = Adafruit_LSM303_Accel_Unified(54321);

/* Assign a unique ID to the mag sensor */
Adafruit_LSM303_Mag_Unified mag = Adafruit_LSM303_Mag_Unified(12345);


void motorCcw(int id);
void motorCw(int id);
void motorBrake(int id);
void motorFloat(int id);
void motorSpeed(int id, int16_t spd);
void restoreCalibration(void);
void saveCalibration(void);


const int leftA2     = 2;
const int leftPwm    = 3;
const int leftA1     = 4;
const int rightA2    = 5;
const int rightPwm   = 6;
const int rightA1    = 7;
const int sensorPin  = A0;
const int linePin    = A1;

const int a1Pins[2]  = {leftA1, rightA1};
const int a2Pins[2]  = {leftA2, rightA2};
const int pwmPins[2] = {leftPwm, rightPwm};

int16_t pwmSpeed = 0;
unsigned long now;
uint16_t verbose = 0x0000;
float wantedHeading = 90.0;
float lastError = 0.0;
float P, I, D, steer;
float kp = 0.0;
float ki = 0.0;
float kd = 0.0;
bool  calibrate = false;

float accBiasX = 0.0;
float accGainX = 1.0;
float accBiasY = 0.0;
float accGainY = 1.0;
float accBiasZ = 0.0;
float accGainZ = 1.0;

float magBiasX = 0.0;
float magBiasY = 0.0;
float magBiasZ = 0.0;
float magGainX = 1.0;
float magGainY = 1.0;
float magGainZ = 1.0;

vector_t accMax;
vector_t accMin;
vector_t magMax;
vector_t magMin;

vector_t accV;
vector_t accR;
vector_t magV;
vector_t magR;


unsigned long target;


// Command parser variables 
char     cmd[40];
int      icmd;



//*************************************************************************
// Basic Arduino setup routine. Gets called once on powerup
//*************************************************************************
void setup() 
{

  // Motor controller pins
  pinMode(leftA2,  OUTPUT);
  pinMode(leftA1,  OUTPUT);
  pinMode(rightA2, OUTPUT);
  pinMode(rightA1, OUTPUT);

  // For now the robot is stationary
  pwmSpeed = 0;

  // Start motor forward
  motorCcw(0);
  motorCcw(1);
  motorSpeed(0, pwmSpeed);
  motorSpeed(1, pwmSpeed);

  // Set up serial port
  Serial.begin(115200);

  Serial.println("");
  Serial.println("");
  Serial.println("Starting");

  /* Initialise the accel sensor */
  if(!acc.begin())
  {
    /* There was a problem detecting the ADXL345 ... check your connections */
    Serial.println("Ooops, no LSM303 detected ... Check your wiring!");
    while(1);
  }

  /* Initialise the mag sensor */
  if(!mag.begin())
  {
    /* There was a problem detecting the LSM303 ... check your connections */
    Serial.println("Ooops, no LSM303 detected ... Check your wiring!");
    while(1);
  }

  // Set mag gain
  mag.setMagGain(LSM303_MAGGAIN_1_3);

  // Restore the last calibration data from EEPROM
  restoreCalibration();

}


//*************************************************************************
// Basic Arduino loop routines. Gets called from an infinite loop.
//*************************************************************************
void loop(void) 
{
  double hdg;
  double accVar, magVar;
  float  error;
  float  errorDot;

  // Get current time
  now = millis();
  
  // Run command parser all the time
  commandParser();

  // Wait for the right time to run the PID loop to follow mag heading
  if (now >= target)
  {
    
    // Get heading
    getHeading(&hdg, &accVar, &magVar);

    // Ignore if bad results
    if (accVar >= 10000.0) return;
    if (magVar >= 10000.0) return;

    // If we got a valid heading, update target time
    target = target + 200;  // Set new target time

    /* Display the results (acceleration is measured in m/s^2) */
    if (verbose & 0x0002) 
    {
        if (calibrate) {
          Serial.print("CAL");
        } else {
          Serial.print("RUN");
        }
                                   Serial.print(",");    
        Serial.print(accV.x);      Serial.print(",");    
        Serial.print(accV.y);      Serial.print(",");    
        Serial.print(accV.z);      Serial.print(",");    
        Serial.print(magV.x);      Serial.print(",");    
        Serial.print(magV.y);      Serial.print(",");    
        Serial.print(magV.z);      Serial.print(",");    
        Serial.print(hdg);         Serial.print(",");    
        Serial.print(accVar);      Serial.print(",");    
        Serial.print(magVar);      Serial.print(",");    
        Serial.println("");
    }
    

    // Determine heading error
    error = hdg - wantedHeading;
    if (error > +180) error -= 360;
    if (error < -180) error += 360;

    // Determine error rate
    errorDot = (error - lastError)/(100.0/1000.0);
    lastError = error;

    #define PID_LIM 150

    // Perform PID loop
    P = error * kp;
    P = constrain(P, -PID_LIM, +PID_LIM);
    I = I + error * ki;
    I = constrain(I, -PID_LIM, +PID_LIM);
    D = errorDot * kd;
    D = constrain(D, -PID_LIM, +PID_LIM);
    steer = P + I + D;
    steer = constrain(steer, -PID_LIM, +PID_LIM);

    // Set motor speeds plus PID adjustment for steering
    motorSpeed(0, pwmSpeed+steer);
    motorSpeed(1, pwmSpeed-steer);

    // Debug data
    if (verbose & 0x0001)
    {
        Serial.print("PID");          Serial.print(", ");
        Serial.print(hdg);            Serial.print(", ");
        Serial.print(wantedHeading);  Serial.print(", ");
        Serial.print(error);          Serial.print(", ");
        Serial.print(errorDot);       Serial.print(", ");
        Serial.print(P);              Serial.print(", ");
        Serial.print(I);              Serial.print(", ");
        Serial.print(D);              Serial.print(", ");
        Serial.print(steer);          Serial.print(", ");
        Serial.print(pwmSpeed+steer); Serial.print(", ");
        Serial.print(pwmSpeed-steer); Serial.print(", ");
        Serial.println("");
    }
    

  }

    
}




//*************************************************************************
// Get heading from magnetometer/accelerometer (in degrees)
//*************************************************************************
void getHeading(double *heading, double *accVar, double *magVar)
{

  double accK = 0.6;
  double magK = 0.6;
  vector_t W;
  vector_t N;
  vector_t from = {1, 0, 0};
  double   x, y;
  double   accT, magT;
  sensors_event_t accEvt;
  sensors_event_t magEvt;

  *magVar = 10000.0;
  *accVar = 10000.0;

  /* Get a new sensor event */
  acc.getEvent(&accEvt);
  mag.getEvent(&magEvt);

  if ( (accEvt.acceleration.x == 0.0) && 
       (accEvt.acceleration.y == 0.0) && 
       (accEvt.acceleration.z == 0.0)) return;

  if ( (magEvt.magnetic.x == 0.0) && 
       (magEvt.magnetic.y == 0.0) && 
       (magEvt.magnetic.z == 0.0)) return;

  /* Calibrate accelerometer */
  accR.x = (accEvt.acceleration.x - accBiasX) * accGainX;
  accR.y = (accEvt.acceleration.y - accBiasY) * accGainY;
  accR.z = (accEvt.acceleration.z - accBiasZ) * accGainZ;
  accT = accR.x*accR.x + accR.y*accR.y + accR.z*accR.z;
  *accVar = (accT - 106.882)/106.882;

  /* filter the accelerometer */
  accV.x = accV.x * accK + accR.x * (1.0-accK);
  accV.y = accV.y * accK + accR.y * (1.0-accK);
  accV.z = accV.z * accK + accR.z * (1.0-accK);
  
  /* Calibrate magnetometer */
  magR.x = (magEvt.magnetic.x - magBiasX) * magGainX;
  magR.y = (magEvt.magnetic.y - magBiasY) * magGainY;
  magR.z = (magEvt.magnetic.z - magBiasZ) * magGainZ;
  magT = magR.x*magR.x + magR.y*magR.y + magR.z*magR.z;
  *magVar = (magT - 1.0)/1.0;
  
  /* filter the magnetometer data */
  magV.x = magV.x * magK + magR.x * (1.0-magK);
  magV.y = magV.y * magK + magR.y * (1.0-magK);
  magV.z = magV.z * magK + magR.z * (1.0-magK);


  // If we are calibrating
  if (calibrate)
  {

    // Capture extremes
    if (accV.x > accMax.x) accMax.x = accV.x;
    if (accV.y > accMax.y) accMax.y = accV.y;
    if (accV.z > accMax.z) accMax.z = accV.z;
    if (accV.x < accMin.x) accMin.x = accV.x;
    if (accV.y < accMin.y) accMin.y = accV.y;
    if (accV.z < accMin.z) accMin.z = accV.z;

    if (magV.x > magMax.x) magMax.x = magV.x;
    if (magV.y > magMax.y) magMax.y = magV.y;
    if (magV.z > magMax.z) magMax.z = magV.z;
    if (magV.x < magMin.x) magMin.x = magV.x;
    if (magV.y < magMin.y) magMin.y = magV.y;
    if (magV.z < magMin.z) magMin.z = magV.z;
    
  }
  
  // Compute W and N vectors
  vector_cross(&magV, &accV, &W);
  W = vector_normalize(&W);
  vector_cross(&accV, &W, &N);
  N = vector_normalize(&N);
    
  // Compute west and north components
  y = vector_dot(&W, &from);
  x = vector_dot(&N, &from);

  // Compute the heading
  *heading = atan2(y, x) * 180.0/3.1415926;
    
  
  return;
}

//*************************************************************************
// Restore calibration from EEPROM
//*************************************************************************
void restoreCalibration(void)
{
        
  // Get cal factors from EEPROM
  EEPROM.get(0,  magBiasX);
  EEPROM.get(4,  magBiasY);
  EEPROM.get(8,  magBiasZ);
  EEPROM.get(12, magGainX);
  EEPROM.get(16, magGainY);
  EEPROM.get(20, magGainZ);
  Serial.println("Calibration restored from EEPROM");
}

//*************************************************************************
// Save calibration to EEPROM
//*************************************************************************
void saveCalibration(void)
{
  // Save cal factors to EEPROM
  EEPROM.put(0,  magBiasX);
  EEPROM.put(4,  magBiasY);
  EEPROM.put(8,  magBiasZ);
  EEPROM.put(12, magGainX);
  EEPROM.put(16, magGainY);
  EEPROM.put(20, magGainZ);
  Serial.println("Calibration values saved to EEPROM");
}


//*************************************************************************
// Run command parser
//*************************************************************************
void commandParser(void) 
{
  char           c;
  char          *p;
  char          *p1;
  char          *p2;


  /* If serial data is available */  
  if (Serial.available() > 0) 
  {
   
    /* Read the next byte */
    c = Serial.read();
   
    /* If the end of a command */
    if ((c == '\n') || (c== '\r')) 
    {
      
      /* Null terminate the command */
      cmd[icmd] = 0;
       
      for(int i = 0; cmd[i]; i++)
      {
        cmd[i] = toupper(cmd[i]);
      }       
     
      /* Parse out the first token */
      p = strtok(cmd, " ");

      /* Test for each command */
       
      /* X */       
      if (strcmp(p, "X") == 0) 
      {
        motorBrake(0);
        motorBrake(1);

      /* p */       
      } else if (strcmp(p, "P") == 0) 
      {         
          p1 = strtok(NULL, " ");
          kp = atof(p1);
          Serial.println(kp);
                  
      /* i */       
      } else if (strcmp(p, "I") == 0) 
      {         
          p1 = strtok(NULL, " ");
          ki = atof(p1);
          Serial.println(ki);
        
      /* d */       
      } else if (strcmp(p, "D") == 0) 
      {         
          p1 = strtok(NULL, " ");
          kd = atof(p1);
          Serial.println(kd);
        
      /* h */       
      } else if (strcmp(p, "H") == 0) 
      {         
          p1 = strtok(NULL, " ");
          wantedHeading = atof(p1);
          Serial.println(wantedHeading);
        
      /* s */       
      } else if (strcmp(p, "S") == 0) 
      {         
          p1 = strtok(NULL, " ");
          pwmSpeed = atoi(p1);
          Serial.println(pwmSpeed);


      /* v */       
      } else if (strcmp(p, "V") == 0) 
      {

          p1 = strtok(NULL, " ");
          verbose = atoi(p1);
          Serial.println(verbose);
         
      /* { */       
      } else if (strcmp(p, "{") == 0) 
      {
        saveCalibration();
         
      /* } */       
      } else if (strcmp(p, "}") == 0) 
      {
        restoreCalibration();

      /* c */       
      } else if (strcmp(p, "C") == 0) 
      {

         // If we are in the calibration mode
         if (calibrate) 
         {
          
           // Turn off calibration flag
           calibrate = false;

           // Compute gains and offsets from captured extremes
           magBiasX = (magMax.x + magMin.x)/2.0;
           magBiasY = (magMax.y + magMin.y)/2.0;
           magBiasZ = (magMax.z + magMin.z)/2.0;
           magGainX = 2.0/(magMax.x - magMin.x);
           magGainY = 2.0/(magMax.y - magMin.y);
           magGainZ = 2.0/(magMax.z - magMin.z);

           // Turn off all printouts to indicate finish
           verbose = 0x000;

         } else {

           // Turn on calibration
           calibrate = true;

           // Set initial extremes
           magMax.x = -1000.0;
           magMax.y = -1000.0;
           magMax.z = -1000.0;
           magMin.x = +1000.0;
           magMin.y = +1000.0;
           magMin.z = +1000.0;

           // Nullify calibration
           magBiasX = 0.0;
           magBiasY = 0.0;
           magBiasZ = 0.0;
           magGainX = 1.0;
           magGainY = 1.0;
           magGainZ = 1.0;

         }
         
      /* ? */       
      } else if (strcmp(p, "?") == 0) 
      {
         Serial.println("Usage:");
         Serial.println(" P n       - Set kP");
         Serial.println(" I n       - Set kI");
         Serial.println(" D n       - Set kD");
         Serial.println(" X         - Panic stop");
         Serial.println(" H n       - Set heading");
         Serial.println(" S n       - Set speed");
         Serial.println(" V n       - Set verbose mode");
         Serial.println(" {         - Save cal to EEPROM");
         Serial.println(" }         - Restore cal from EEPROM");
         Serial.println(" C         - Toggle calibration");
         Serial.println(" !         - Print info");
         
      /* ! */       
      } else if (strcmp(p, "!") == 0) 
      {
         Serial.print("S, ");             Serial.println(pwmSpeed);
         Serial.print(" kp,");            Serial.println(kp);
         Serial.print(" ki,");            Serial.println(ki);
         Serial.print(" kd,");            Serial.println(kd);
         Serial.print(" wantedHeading,"); Serial.println(wantedHeading);
         Serial.print(" magBiasX,");      Serial.println(magBiasX);
         Serial.print(" magBiasY,");      Serial.println(magBiasY);
         Serial.print(" magBiasZ,");      Serial.println(magBiasZ);
         Serial.print(" magGainX,");      Serial.println(magGainX);
         Serial.print(" magGainY,");      Serial.println(magGainY);
         Serial.print(" magGainZ,");      Serial.println(magGainZ);
         Serial.println("");
         
         
      /* Not a valid command */  
      } else {
         Serial.print("Bad command (");
         Serial.print(p);
         Serial.println(")");
      }
     
       /* Now that we parsed the command, reset for next command */
       icmd=0;

     /* If this is not a terminator */     
     } else 
     {
       
       /* Append the character to the command buffer */
       cmd[icmd] = c;
       icmd++;
       if (icmd >= 80) {
         Serial.println("Command buffer overflow");
         icmd = 0;
       }
     } /* end terminator */
   } /* end character available */
}


//*************************************************************************
/* Set the pins for CCW motion */
//*************************************************************************
void motorCcw(int id)
{
  digitalWrite(a1Pins[id], 0);
  digitalWrite(a2Pins[id], 1);
}

//*************************************************************************
/* Set the pins for CW motion */
//*************************************************************************
void motorCw(int id)
{
  digitalWrite(a1Pins[id], 1);
  digitalWrite(a2Pins[id], 0);
}

//*************************************************************************
/* Set the pins for brake action */
//*************************************************************************
void motorBrake(int id)
{
  digitalWrite(a1Pins[id], 1);
  digitalWrite(a2Pins[id], 1);
}

//*************************************************************************
/* Set the pins for float action */
//*************************************************************************
void motorFloat(int id)
{
  digitalWrite(a1Pins[id], 0);
  digitalWrite(a2Pins[id], 0);
}

//*************************************************************************
// Set the motor speed and direction
//*************************************************************************
void motorSpeed(int id, int16_t spd)
{
  
  if (spd < 0)
  {
      motorCcw(id);
      spd = -spd;
  } else {
      motorCw(id);
  }
  
  // Dead band on motor so it does not hum with small speeds
  if (spd < 20) spd = 0.0;

  // Limit to max PWM
  if (spd > 255) spd = 255;
  
  analogWrite(pwmPins[id], spd);         
}


